"""
Entry point for the BoxJelly application.
"""

from pathlib import Path
import sys

from PyQt6 import QtWidgets, QtGui, QtCore

import boxjelly.lib.resources
from boxjelly.lib.constants import APP_NAME, APP_ORGANIZATION, CONTROL_PORT_DEFAULT, FRAME_RATE_DEFAULT, INCOMING_PORT_DEFAULT, INCOMING_TOPIC_DEFAULT, OUTGOING_PORT_DEFAULT, OUTGOING_TOPIC_DEFAULT
from boxjelly.lib.settings import SettingsManager
from boxjelly.ui.MainWindow import MainWindow


def start(argv):
    """Start the application"""
    # Create the application
    app = QtWidgets.QApplication(argv)
    
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.Format.IniFormat)
    init_settings()
    
    app.setApplicationName(APP_NAME)
    app.setOrganizationName(APP_ORGANIZATION)
    
    # Create the splash screen
    splash_logo = QtGui.QIcon(':/icons/logo').pixmap(256, 256)
    splash = QtWidgets.QSplashScreen(splash_logo)
    splash.show()
    
    # Create the main window
    window = MainWindow()
    window.show()
    splash.finish(window)
    
    # Start the application
    app.exec()


def init_settings():
    """Initialize settings"""
    settings = SettingsManager.get_instance()
    
    settings.control_port = ('network/control_port', int, CONTROL_PORT_DEFAULT)
    settings.incoming_port = ('network/incoming_port', int, INCOMING_PORT_DEFAULT)
    settings.incoming_topic = ('network/incoming_topic', str, INCOMING_TOPIC_DEFAULT)
    settings.outgoing_port = ('network/outgoing_port', int, OUTGOING_PORT_DEFAULT)
    settings.outgoing_topic = ('network/outgoing_topic', str, OUTGOING_TOPIC_DEFAULT)
    settings.selection_enabled = ('network/selection_enabled', bool, False)
    
    settings.frame_rate = ('video/frame_rate', float, FRAME_RATE_DEFAULT)
    


def main():
    """Parse command line arguments and start"""
    start(sys.argv)


if __name__ == "__main__":
    main()



from PyQt6 import QtWidgets

from boxjelly.ui.settings.tabs.AbstractSettingsTab import AbstractSettingsTab


class NetworkSettingsTab(AbstractSettingsTab):
    """
    Network settings tab.
    """
    
    def __init__(self, parent=None):
        super().__init__('Network', parent=parent)
        
        self._remote_control_section = QtWidgets.QGroupBox('Remote Control')
        
        self._control_port_edit = QtWidgets.QSpinBox()
        self._control_port_edit.setRange(0, 65535)
        self._control_port_edit.setValue(self._settings.control_port.value)
        
        self._localization_section = QtWidgets.QGroupBox('Localization')
        
        self._incoming_port_edit = QtWidgets.QSpinBox()
        self._incoming_port_edit.setRange(0, 65535)
        self._incoming_port_edit.setValue(self._settings.incoming_port.value)
        
        self._incoming_port_topic_edit = QtWidgets.QLineEdit()
        self._incoming_port_topic_edit.setText(self._settings.incoming_topic.value)
        
        self._outgoing_port_edit = QtWidgets.QSpinBox()
        self._outgoing_port_edit.setRange(0, 65535)
        self._outgoing_port_edit.setValue(self._settings.outgoing_port.value)
        
        self._outgoing_port_topic_edit = QtWidgets.QLineEdit()
        self._outgoing_port_topic_edit.setText(self._settings.outgoing_topic.value)
        
        self._cthulhu_section = QtWidgets.QGroupBox('Cthulhu')
        self._selection_enabled_checkbox = QtWidgets.QCheckBox()
        self._selection_enabled_checkbox.setChecked(self._settings.selection_enabled.value)
        
        self._control_port_edit.valueChanged.connect(self._settings_changed)
        self._incoming_port_edit.valueChanged.connect(self._settings_changed)
        self._incoming_port_topic_edit.textChanged.connect(self._settings_changed)
        self._outgoing_port_edit.valueChanged.connect(self._settings_changed)
        self._outgoing_port_topic_edit.textChanged.connect(self._settings_changed)
        self._selection_enabled_checkbox.stateChanged.connect(self._settings_changed)
        
        self._arrange()
    
    def _arrange(self):
        layout = QtWidgets.QVBoxLayout()
        
        remote_control_layout = QtWidgets.QFormLayout()
        remote_control_layout.addRow('Control Port', self._control_port_edit)
        self._remote_control_section.setLayout(remote_control_layout)
        
        localization_layout = QtWidgets.QFormLayout()
        localization_layout.addRow('Incoming Port', self._incoming_port_edit)
        localization_layout.addRow('Incoming Topic', self._incoming_port_topic_edit)
        localization_layout.addRow('Outgoing Port', self._outgoing_port_edit)
        localization_layout.addRow('Outgoing Topic', self._outgoing_port_topic_edit)
        self._localization_section.setLayout(localization_layout)
        
        cthulhu_layout = QtWidgets.QFormLayout()
        cthulhu_layout.addRow('Enable Selection Highlight', self._selection_enabled_checkbox)
        self._cthulhu_section.setLayout(cthulhu_layout)
        
        layout.addWidget(self._remote_control_section)
        layout.addWidget(self._localization_section)
        layout.addWidget(self._cthulhu_section)
        
        self.setLayout(layout)
    
    def apply_settings(self):
        # Remote control
        self._settings.control_port.value = self._control_port_edit.value()
        
        # Localization control
        if self._settings.incoming_port.value != self._incoming_port_edit.value():
            self._settings.incoming_port.value = self._incoming_port_edit.value()
        
        if self._settings.incoming_topic.value != self._incoming_port_topic_edit.text():
            self._settings.incoming_topic.value = self._incoming_port_topic_edit.text()
        
        if self._settings.outgoing_port.value != self._outgoing_port_edit.value():
            self._settings.outgoing_port.value = self._outgoing_port_edit.value()
        
        if self._settings.outgoing_topic.value != self._outgoing_port_topic_edit.text():
            self._settings.outgoing_topic.value = self._outgoing_port_topic_edit.text()
        
        # Selection enable/disable
        self._settings.selection_enabled.value = self._selection_enabled_checkbox.isChecked()
    


from PyQt6 import QtWidgets, QtGui, QtCore

from boxjelly.models.TrackListModel import TrackListModel
from boxjelly.ui.graphicsitems.TimelineTrack import TimelineTrack

class TrackGraphicsView(QtWidgets.QGraphicsView):
    
    TRACK_HEIGHT = 30
    
    frameSelected = QtCore.pyqtSignal(int)
    scaleChanged = QtCore.pyqtSignal(float, float)
    
    verticalScrollBarChanged = QtCore.pyqtSignal(int)
    horizontalScrollBarChanged = QtCore.pyqtSignal(int)
    frameScrolled = QtCore.pyqtSignal(int)
    
    def __init__(self, parent, track_model: TrackListModel, selection_model: QtCore.QItemSelectionModel):
        super().__init__(parent)
        
        self._track_model = track_model
        self._track_model.dataChanged.connect(self._on_track_model_data_changed)
        self._track_model.modelReset.connect(self._on_track_model_reset)
        self._track_model.rowsInserted.connect(self._on_track_model_rows_inserted)
        self._track_model.rowsAboutToBeRemoved.connect(self._on_track_model_rows_about_to_be_removed)
        self._track_model.rowsRemoved.connect(self._on_track_model_rows_removed)
        self._track_model.layoutChanged.connect(self._on_track_model_layout_changed)
        
        self._selection_model = selection_model
        
        self._scene = QtWidgets.QGraphicsScene()
        self.setScene(self._scene)
        self.setViewportMargins(0, 0, 0, 0)
        self.setFrameStyle(QtWidgets.QFrame.Shape.NoFrame)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignTop)
        
        self._background_rect = None  # backround rectangle
        self._timeline_track_dict = {}  # track ID -> TimelineTrack graphics item
        
        self._current_frame = 0  # current frame that the playhead is on
        self._frame_scale = 1.0  # horizontal scale of the timeline, in scene units per frame
        
        self.scaleChanged.connect(self._rescale)
        
        self.set_frame_scale(self._frame_scale)
        
        self.verticalScrollBar().valueChanged.connect(self.verticalScrollBarChanged.emit)
        self.horizontalScrollBar().valueChanged.connect(self.horizontalScrollBarChanged.emit)
        self.horizontalScrollBar().valueChanged.connect(lambda value: self.frameScrolled.emit(self._map_scene_to_frame(value)))
    
    def compute_frame_count(self):
        if self._track_model.rowCount(None) == 0:
            return 0
    
        end_frames = [self._track_model.data(self._track_model.index(row_idx, 0), TrackListModel.EndFrameRole) for row_idx in range(self._track_model.rowCount(None))]
        return max(end_frames) + 1
    
    def is_loaded(self):
        return self.compute_frame_count() > 0
    
    @QtCore.pyqtSlot(QtCore.QModelIndex, QtCore.QModelIndex)
    def _on_track_model_data_changed(self, top_left: QtCore.QModelIndex, bottom_right: QtCore.QModelIndex):
        """Handle track data changes."""
        if not self.is_loaded():
            return
        
        for row_idx in range(top_left.row(), bottom_right.row() + 1):
            index = self._track_model.index(row_idx, 0)
            track_id = self._track_model.data(index, TrackListModel.IDRole)
            
            timeline_track = self._timeline_track_dict[track_id]
            timeline_track.update()
    
    @QtCore.pyqtSlot()
    def _on_track_model_reset(self):
        """When the track model is reset, reset the scene."""
        self._reset_scene()
    
    @QtCore.pyqtSlot(QtCore.QModelIndex, int, int)
    def _on_track_model_rows_inserted(self, parent, start, end):
        """When tracks are added, add new timeline tracks."""
        for row_idx in range(start, end + 1):
            track = self._track_model.get_track(row_idx)
            
            timeline_track = TimelineTrack(track)
            
            timeline_track.setZValue(0)
            timeline_track.set_scale(self._frame_scale, 1.0)
            self.scaleChanged.connect(timeline_track.set_scale)
            timeline_track.clicked.connect(self._on_timeline_track_clicked)
            
            self._scene.addItem(timeline_track)
            self._timeline_track_dict[track.id] = timeline_track
        
        # Trigger a rescale
        self._rescale(self._frame_scale, ...)
    
    @QtCore.pyqtSlot(object)
    def _on_timeline_track_clicked(self, timeline_track: TimelineTrack):
        """When timeline track is clicked, select it."""
        model_index = self._track_model.index(self._track_model.index_by_id(timeline_track.track.id), 0)
        self._selection_model.select(model_index, QtCore.QItemSelectionModel.SelectionFlag.ClearAndSelect)
    
    @QtCore.pyqtSlot('QList<QPersistentModelIndex>', QtCore.QAbstractItemModel.LayoutChangeHint)
    def _on_track_model_layout_changed(self, removed_indexes, change_hint):
        """When the track model layout changes, reposition the tracks."""
        self._reposition_tracks()
    
    @QtCore.pyqtSlot(QtCore.QModelIndex, int, int)
    def _on_track_model_rows_about_to_be_removed(self, parent, start, end):
        """When tracks are removed, remove the corresponding timeline tracks."""
        for row_idx in range(start, end + 1):
            track = self._track_model.get_track(row_idx)
            
            timeline_track = self._timeline_track_dict[track.id]
            
            self._scene.removeItem(timeline_track)
            del self._timeline_track_dict[track.id]
    
    @QtCore.pyqtSlot(QtCore.QModelIndex, int, int)
    def _on_track_model_rows_removed(self, parent, start, end):
        self._reposition_tracks()
    
    def _center_on_frame(self, frame: int):
        self.horizontalScrollBar().setValue(self._map_frame_to_scene(frame) - self.width() / 2)
    
    def set_frame_scale(self, scale: float):
        self._frame_scale = scale
        self._center_on_frame(self._current_frame)
        self.scaleChanged.emit(scale, 1.0)
    
    def _map_scene_to_frame(self, scene_x: float) -> int:
        return int(scene_x / self._frame_scale)
    
    def _map_frame_to_scene(self, frame: int) -> float:
        return frame * self._frame_scale
    
    @property
    def scene_width(self) -> float:
        if not self.is_loaded():
            return 0.
        
        return self._map_frame_to_scene(self.compute_frame_count())

    @property
    def scene_height(self):
        if not self.is_loaded():
            return 0.
        
        return self.TRACK_HEIGHT * len(self._track_model)
    
    @QtCore.pyqtSlot(float, float)
    def _rescale(self, new_frame_scale: float, _):
        """
        Rescale scene.
        """
        if not self.is_loaded():
            return
        
        self._frame_scale = new_frame_scale
        
        self._reposition_tracks()
        
        self._scene.setSceneRect(0, 0, self.scene_width, self.TRACK_HEIGHT * self._track_model.rowCount(None))  # update scene size
        
        self._background_rect.setRect(0, 0, self._scene.width(), self._scene.height())  # rescale background rect
    
    def _clear_scene(self):
        self._scene.clear()
        
        self._background_rect = None
        self._timeline_track_dict = {}
    
    def _reset_scene(self):
        self._clear_scene()
        self._add_background()
        self._reposition_tracks()
        self.set_frame_scale(self._frame_scale)
    
    def _add_background(self):
        """Add a background rectangle to the scene."""
        self._background_rect = QtWidgets.QGraphicsRectItem(0, 0, self._scene.width(), self._scene.height())
        self._background_rect.setBrush(QtGui.QColor(0, 0, 0, 0))
        self._background_rect.setPen(QtGui.QPen(QtCore.Qt.PenStyle.NoPen))
        self._background_rect.setZValue(-1)
        
        self._scene.addItem(self._background_rect)
    
    def _reposition_tracks(self):
        """Reposition tracks."""
        for row_idx in range(self._track_model.rowCount(None)):
            index = self._track_model.index(row_idx, 0)
            track_id = self._track_model.data(index, TrackListModel.IDRole)
            
            timeline_track = self._timeline_track_dict[track_id]
            timeline_track.setPos(timeline_track.track.start_frame * self._frame_scale, row_idx * self.TRACK_HEIGHT)
    
    def show_frame(self, frame: int):
        """Update the view to a specific frame, trying to center it in the viewport."""
        if not self.is_loaded():
            return
        
        self._current_frame = frame
        
        new_scrollbar_position = self._map_frame_to_scene(frame) - self.viewport().width() / 2
        if new_scrollbar_position < 0:
            new_scrollbar_position = 0
        elif new_scrollbar_position > self.scene_width - self.viewport().width():
            new_scrollbar_position = self.scene_width - self.viewport().width()
        self.horizontalScrollBar().setValue(new_scrollbar_position)
        
        self.viewport().repaint()  # Trigger a repaint of the viewport.
    
    def paintEvent(self, event: QtGui.QPaintEvent):
        super().paintEvent(event)
        
        if not self.is_loaded():
            return
        
        # Draw a red line to mark the current frame.
        painter = QtGui.QPainter(self.viewport())
        scene_x = self._map_frame_to_scene(self._current_frame)
        view_x = self.mapFromScene(scene_x, 0).x()
        painter.setPen(QtGui.QPen(QtGui.QColor(255, 0, 0), 1))
        painter.drawLine(view_x, 0, view_x, self._scene.height())
    
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        """Handle mouse press events."""
        super().mousePressEvent(event)
        
        scene_pos = self.mapToScene(event.pos())
        
        if event.button() == QtCore.Qt.MouseButton.LeftButton:  # Left click
            self.frameSelected.emit(self._map_scene_to_frame(scene_pos.x()))  # Emit the frame selected signal.
    
    def wheelEvent(self, event: QtGui.QWheelEvent):
        """Handle mouse wheel events."""
        
        # If Ctrl+scroll, change the frame scale.
        if event.modifiers() == QtCore.Qt.KeyboardModifier.ControlModifier:
            if event.angleDelta().y() > 0:
                self.set_frame_scale(round(self._frame_scale * 1.1, 2))
            else:
                self.set_frame_scale(round(self._frame_scale / 1.1, 2))
            return
        
        # Otherwise, scroll the view.
        super().wheelEvent(event)
    


"""
Track-related models.
"""

from typing import List, Optional
from dataclasses import dataclass
from uuid import UUID
from statistics import mode

from dataclasses_json import dataclass_json

@dataclass_json
@dataclass
class BoundingBox:
    x: int
    y: int
    w: int
    h: int


@dataclass_json
@dataclass
class VisualEvent:
    id: UUID
    box: BoundingBox
    label: str
    confidence: float


@dataclass_json
@dataclass
class Track:
    id: UUID
    start_frame: int
    events: List[Optional[VisualEvent]]
    
    def __len__(self) -> int:
        return len(self.events)

    @property
    def label_mode(self) -> str:
        return mode(event.label for event in self.events if event is not None)
    
    @property
    def slice(self):
        return slice(self.start_frame, self.start_frame + len(self))



from PyQt6 import QtWidgets, QtCore

from boxjelly.ui.WidgetComboBox import WidgetComboBox
from boxjelly.ui.SourceWidgets import LocalVideoMediaSourceWidget, URLMediaSourceWidget


class MediaSelectionBox(QtWidgets.QGroupBox):
    """Media selection group box"""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        
        self.setTitle('Media')
        
        self.widget_combo_box = WidgetComboBox(self)
        self.widget_combo_box.add_widget(LocalVideoMediaSourceWidget(self), 'Video file')
        self.widget_combo_box.add_widget(URLMediaSourceWidget(self), 'Video URL')
        self.widget_combo_box.refresh()
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Add widget combo box
        self.layout().addWidget(self.widget_combo_box)
        
        # Add stretch
        self.layout().addStretch()
    
    @property
    def media_content(self) -> QtCore.QUrl:
        return self.widget_combo_box.widget.media_content



from typing import Optional

from PyQt6 import QtWidgets, QtCore, QtGui

from boxjelly.lib.settings import SettingsManager


class AbstractSettingsTab(QtWidgets.QWidget):
    """
    Abstract settings tab.
    
    Contains a page of settings for the application.
    """
    
    settingsChanged = QtCore.pyqtSignal()
    
    def __init__(self, name: str, icon: Optional[QtGui.QIcon] = None, parent=None):
        super().__init__(parent)
        
        self._settings = SettingsManager.get_instance()
        
        self._name = name
        self._icon = icon
    
    @property
    def name(self):
        return self._name
    
    @property
    def icon(self) -> Optional[QtGui.QIcon]:
        return None
    
    def apply_settings(self):
        raise NotImplementedError()
    
    def _settings_changed(self):
        self.settingsChanged.emit()



from PyQt6 import QtWidgets, QtMultimedia, QtCore

from boxjelly.lib.fileio import AbstractTrackFileIO, DeepseaTrackFileIO, JSONTrackFileIO, YOLOv5DeepSortTrackFileIO
from boxjelly.ui.FileSelector import FileSelector


class AbstractMediaSourceWidget(QtWidgets.QWidget):
    """Abstract media source widget"""
    
    @property
    def media_content(self) -> QtCore.QUrl:
        raise NotImplementedError()


class AbstractTrackSourceWidget(QtWidgets.QWidget):
    """Abstract track source widget"""
    
    @property
    def io(self) -> AbstractTrackFileIO:
        raise NotImplementedError()


class LocalVideoMediaSourceWidget(AbstractMediaSourceWidget):
    """Video file media source widget. Loads local video files."""
    def __init__(self, parent):
        super().__init__(parent)
    
        self._filename = None
        
        self._selector = FileSelector(self, filter='Video files (*.mp4 *.mov *.wmv *.avi)')
        self._selector.fileSelected.connect(self._update_filename)

        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange the file selector
        self.layout().addWidget(self._selector)
    
    def _update_filename(self, filename):
        self._filename = filename
    
    @property
    def media_content(self) -> QtCore.QUrl:
        return QtCore.QUrl.fromLocalFile(self._filename)


class URLMediaSourceWidget(AbstractMediaSourceWidget):
    """URL media source widget. Loads video files from URL."""
    def __init__(self, parent):
        super().__init__(parent)
    
        self._url = None
        
        self._line_edit = QtWidgets.QLineEdit()
        self._line_edit.setPlaceholderText('URL')
        self._line_edit.textChanged.connect(self._update_url)
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange the text input
        self.layout().addWidget(self._line_edit)
    
    def _update_url(self, url):
        self._url = url
    
    @property
    def media_content(self) -> QtCore.QUrl:
        return QtCore.QUrl(self._url)


class YOLOv5DeepSortTrackSourceWidget(AbstractTrackSourceWidget):
    """YOLOv5-DeepSort file track source widget. Loads tracks from a modified MOT text file."""
    def __init__(self, parent):
        super().__init__(parent)
    
        self._filename = None
        
        self._selector = FileSelector(self, filter='Text files (*.txt)')
        self._selector.fileSelected.connect(self._update_filename)
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange the file selector
        self.layout().addWidget(self._selector)
    
    def _update_filename(self, filename):
        self._filename = filename
    
    @property
    def io(self) -> YOLOv5DeepSortTrackFileIO:
        return YOLOv5DeepSortTrackFileIO(self._filename)


class DeepseaTrackSourceWidget(AbstractTrackSourceWidget):
    """deepsea-track artifact track source widget. Loads tracks from a tar.gz artifact of track JSONs as output by deepsea-track."""
    def __init__(self, parent):
        super().__init__(parent)
        
        self._filename = None
        
        self._selector = FileSelector(self, filter='Tar.gz files (*.tar.gz)')
        self._selector.fileSelected.connect(self._update_filename)
        
        self.arrange()
        
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange the file selector
        self.layout().addWidget(self._selector)
    
    def _update_filename(self, filename):
        self._filename = filename
    
    @property
    def io(self) -> DeepseaTrackFileIO:
        return DeepseaTrackFileIO(self._filename)


class JSONTrackSourceWidget(AbstractTrackSourceWidget):
    """JSON track source widget"""
    def __init__(self, parent):
        super().__init__(parent)
        
        self._filename = None
        
        self._selector = FileSelector(self, filter='JSON files (*.json)')
        self._selector.fileSelected.connect(self._update_filename)
        
        self.arrange()
        
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Set zero margins
        self.layout().setContentsMargins(0, 0, 0, 0)
        
        # Arrange the file selector
        self.layout().addWidget(self._selector)
    
    def _update_filename(self, filename):
        self._filename = filename
    
    @property
    def io(self) -> JSONTrackFileIO:
        return JSONTrackFileIO(self._filename)
        


from PyQt6 import QtWidgets

from boxjelly.ui.WidgetComboBox import WidgetComboBox
from boxjelly.ui.SourceWidgets import DeepseaTrackSourceWidget, YOLOv5DeepSortTrackSourceWidget, JSONTrackSourceWidget
from boxjelly.lib.fileio import AbstractTrackFileIO


class TrackSelectionBox(QtWidgets.QGroupBox):
    """Track selection group box"""
    
    def __init__(self, parent=None):
        super().__init__(parent)
        
        self.setTitle('Tracks')
        
        self.widget_combo_box = WidgetComboBox(self)
        self.widget_combo_box.add_widget(YOLOv5DeepSortTrackSourceWidget(self), 'YOLOv5-DeepSort')
        self.widget_combo_box.add_widget(DeepseaTrackSourceWidget(self), 'deepsea-track')
        self.widget_combo_box.add_widget(JSONTrackSourceWidget(self), 'JSON')
        self.widget_combo_box.refresh()
        
        self.arrange()
    
    def arrange(self):
        """Arrange the widget"""
        # Set vertical layout
        self.setLayout(QtWidgets.QVBoxLayout())
        
        # Add widget combo box
        self.layout().addWidget(self.widget_combo_box)
        
        # Add stretch
        self.layout().addStretch()
    
    @property
    def io(self) -> AbstractTrackFileIO:
        return self.widget_combo_box.widget.io


from PyQt6 import QtCore, QtMultimedia, QtWidgets, QtGui

from boxjelly.commands.TrackListCommand import BatchRenameTracksCommand, DeleteTrackByIDCommand, BatchDeleteTracksByIDCommand, BatchRenameTracksCommand, SplitTrackCommand, MergeTracksCommand
from boxjelly.lib.constants import APP_NAME
from boxjelly.models.TrackListModel import TrackListModel
from boxjelly.ui.OpenDialog import OpenDialog
from boxjelly.ui.settings.SettingsDialog import SettingsDialog
from boxjelly.ui.track.TrackPanel import TrackPanel
from boxjelly.ui.video.CthulhuVideoController import CthulhuVideoController

class MainWindow(QtWidgets.QMainWindow):
    """Application main window"""
    
    mediaLoaded = QtCore.pyqtSignal(QtCore.QUrl)
    tracksLoaded = QtCore.pyqtSignal(list)
    
    readRequested = QtCore.pyqtSignal()
    writeRequested = QtCore.pyqtSignal(list)
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        # Set window title
        self.setWindowTitle(APP_NAME)
        
        # Set window icon
        self.setWindowIcon(QtGui.QIcon(':/icons/logo'))
        
        # Create media
        self._media = None
        
        # Create track list model
        self._track_model = TrackListModel(self)
        
        # Create selection model
        self._selection_model = QtCore.QItemSelectionModel(self._track_model)
        
        # Create undo stack
        self._undo_stack = QtGui.QUndoStack(self)
        
        # Create the menu bar
        self._create_menu_bar()
        
        # Create the track panel and set it as the main widget
        self._track_panel = TrackPanel(self, self._track_model, self._selection_model)
        self.setCentralWidget(self._track_panel)
        
        # Create the Cthulhu video controller
        self._video_controller = CthulhuVideoController(self, self._track_model, self._selection_model)
        
        # Create the settings dialog
        self._settings_dialog = SettingsDialog(self)
        
        # Media content and track IO
        self._media_content = None
        self._track_io = None
        
        # Create the background worker thread
        self._worker_thread = QtCore.QThread()
        self._worker_thread.start()
        
        # Connect signals
        self.mediaLoaded.connect(self._video_controller.set_media)
        self._video_controller.mediaSetStatus.connect(self._on_media_set_status)
        self._track_panel.frameSelected.connect(self._video_controller.set_frame)
        self._video_controller.frameUpdated.connect(self._track_panel.show_frame)
        self._selection_model.selectionChanged.connect(self._on_selection_changed)
    
    @QtCore.pyqtSlot(bool)
    def _on_media_set_status(self, status):
        """Handle the media set status."""
        if not status:  # Failure to set media
            QtWidgets.QMessageBox.critical(self, 'Error', 'Failed to open media in Cthulhu.\nIs Cthulhu running?')
    
    @QtCore.pyqtSlot(QtCore.QItemSelection, QtCore.QItemSelection)
    def _on_selection_changed(self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection):
        # Select
        for model_index in selected.indexes():
            self._track_model.select_track(model_index.row())
        
        # Deselect
        for model_index in deselected.indexes():
            self._track_model.deselect_track(model_index.row())

    def _create_menu_bar(self):
        menu_bar = self.menuBar()
        
        # File menu
        file_menu = menu_bar.addMenu('&File')
        file_menu.addAction('&Open', QtGui.QKeySequence('Ctrl+O'), self.open)
        file_menu.addAction('&Save', QtGui.QKeySequence('Ctrl+S'), self.save)
        file_menu.addAction('Save &As', QtGui.QKeySequence('Ctrl+Shift+S'), self.save_as)
        file_menu.addSeparator()
        file_menu.addAction('Se&ttings', QtGui.QKeySequence('Ctrl+,'), self.open_settings)
        file_menu.addSeparator()
        file_menu.addAction('&Quit', QtGui.QKeySequence('Ctrl+Q'), self.close)
        
        # Edit menu
        edit_menu = menu_bar.addMenu('&Edit')
        rename_action = edit_menu.addAction(QtGui.QIcon(':/icons/rename'), 'Re&name', QtGui.QKeySequence('Ctrl+R'), self.rename_selected_tracks)
        split_action = edit_menu.addAction(QtGui.QIcon(':/icons/split'), '&Split', QtGui.QKeySequence('Ctrl+N'), self.split_selected_track)
        merge_action = edit_menu.addAction(QtGui.QIcon(':/icons/merge'), '&Merge', QtGui.QKeySequence('Ctrl+M'), self.merge_selected_tracks)
        delete_action = edit_menu.addAction(QtGui.QIcon(':/icons/delete'), '&Delete', QtGui.QKeySequence('Del'), self.delete_selected_tracks)
        edit_menu.addSeparator()
        undo_action = edit_menu.addAction(QtGui.QIcon(':/icons/undo'), '&Undo', QtGui.QKeySequence('Ctrl+Z'), self._undo_stack.undo)
        redo_action = edit_menu.addAction(QtGui.QIcon(':/icons/redo'), '&Redo', QtGui.QKeySequence('Ctrl+Shift+Z'), self._undo_stack.redo)
        
        # Go menu
        go_menu = menu_bar.addMenu('&Go')
        seek_beginning_action = go_menu.addAction(QtGui.QIcon(':/icons/seek_beginning'), 'Seek &beginning', QtGui.QKeySequence('Ctrl+['), self.seek_beginning_selected_tracks)
        seek_end_action = go_menu.addAction(QtGui.QIcon(':/icons/seek_end'), 'Seek &end', QtGui.QKeySequence('Ctrl+]'), self.seek_end_selected_tracks)
        
        # Add actions to toolbar
        toolbar = QtWidgets.QToolBar('Edit', self)
        self.addToolBar(QtCore.Qt.ToolBarArea.LeftToolBarArea, toolbar)
        toolbar.addActions([rename_action, split_action, merge_action, delete_action])
        toolbar.addSeparator()
        toolbar.addActions([seek_beginning_action, seek_end_action])
        toolbar.addSeparator()
        toolbar.addActions([undo_action, redo_action])
    
    def open_settings(self):
        """Open the settings dialog."""
        self._settings_dialog.show()
    
    def open(self):
        """Show the open dialog and load."""
        # Create and run the dialog
        open_dialog = OpenDialog(self)
        open_dialog.exec()
        
        # If not accepted, return
        if open_dialog.result() != QtWidgets.QDialog.DialogCode.Accepted:
            return
        
        # Otherwise, get the media content and track IO
        self._media_content = open_dialog.media_content
        self._track_io = open_dialog.track_io
        
        # Load the media
        self.load_media(self._media_content)
        
        # Run the track IO read in the worker thread thread
        self._track_io.moveToThread(self._worker_thread)
        self._track_io.fileRead.connect(self._track_model.set_tracks)
        self._track_io.fileWritten.connect(self._on_save_finished)
        self._track_io.fileReadFailure.connect(self._on_read_failure)
        self._track_io.fileWriteFailure.connect(self._on_write_failure)
        self.readRequested.connect(self._track_io.read)
        self.writeRequested.connect(self._track_io.write)
        self.do_read()
    
    def do_read(self):
        """Read the tracks from the track IO."""
        self.readRequested.emit()
        
        # Show a progress dialog while reading
        progress = QtWidgets.QProgressDialog('Reading tracks...', None, 0, 0, self)
        progress.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        progress.setMinimumDuration(0)
        progress.show()
        
        # Wait for the read to finish or fail, then close
        self._track_io.fileRead.connect(progress.close)
        self._track_io.fileReadFailure.connect(progress.close)
    
    @QtCore.pyqtSlot(str)
    def _on_read_failure(self, error_message: str):
        """Show a message box when reading fails."""
        QtWidgets.QMessageBox.critical(self, 'Error', 'Failed to read tracks: \n\n{}'.format(error_message))
    
    def do_write(self):
        """Write the tracks to the track IO."""
        self.writeRequested.emit(self._track_model.get_all_tracks())
    
    @QtCore.pyqtSlot(str)
    def _on_write_failure(self, error_message: str):
        """Show a message box when writing fails."""
        QtWidgets.QMessageBox.critical(self, 'Error', 'Failed to write tracks: \n\n{}'.format(error_message))
    
    @QtCore.pyqtSlot()
    def _on_save_finished(self):
        QtWidgets.QMessageBox.information(self, 'Saved', 'Saved successfully.')
    
    def save(self):
        # If no track IO, show an error
        if not self._track_io:
            QtWidgets.QMessageBox.warning(self, 'Error', 'No file open.')
            return
        
        # Get confirmation
        confirm = QtWidgets.QMessageBox.question(self, 'Confirm', 'Are you sure you want to save?')
        
        # If confirmed, send the signal to save
        if confirm == QtWidgets.QMessageBox.StandardButton.Yes:
            self.do_write()
    
    def save_as(self):
        # If no track IO, show an error
        if not self._track_io:
            QtWidgets.QMessageBox.warning(self, 'Error', 'No file open.')
            return
        
        # Get the new file path, requiring the same suffix as the current
        current_dir = self._track_io.path.parent
        current_suffix = self._track_io.path.suffix
        new_path_str, ok = QtWidgets.QFileDialog.getSaveFileName(self, 'Save As', str(current_dir), f'{current_suffix[1:].upper()} (*{current_suffix})')
        
        if not ok:
            return
        
        # Update the track IO target path
        self._track_io.path = new_path_str
        
        # Send the signal to save
        self.do_write()
        
    
    @QtCore.pyqtSlot(QtCore.QUrl)
    def load_media(self, media):
        """Load the media and emit the mediaLoaded signal."""
        self._media = media
        self.mediaLoaded.emit(media)
    
    @QtCore.pyqtSlot(list)
    def load_tracks(self, tracks):
        """Load the tracks and emit the tracksLoaded signal."""
        self._track_model.set_tracks(tracks)
        self.tracksLoaded.emit(tracks)
    
    @QtCore.pyqtSlot(int)
    def delete_track(self, track_id):
        """Delete a track from the model by its ID."""
        command = DeleteTrackByIDCommand(self._track_model, track_id)
        self._undo_stack.push(command)
    
    def delete_selected_tracks(self):
        """Delete the selected tracks."""
        selected_indices = self._selection_model.selectedIndexes()
        selected_ids = [index.data(TrackListModel.IDRole) for index in selected_indices]
        
        command = BatchDeleteTracksByIDCommand(self._track_model, selected_ids)
        self._undo_stack.push(command)
    
    def rename_selected_tracks(self):
        """Change the label of a track by its ID."""
        new_label, ok = QtWidgets.QInputDialog.getText(self, 'Rename track', 'Label:')
        if not ok:
            return
        
        selected_indices = self._selection_model.selectedIndexes()
        selected_ids = [index.data(TrackListModel.IDRole) for index in selected_indices]
        
        command = BatchRenameTracksCommand(self._track_model, selected_ids, new_label)
        self._undo_stack.push(command)
        
    def split_selected_track(self):
        """Split a track at the current frame."""
        selected_indices = self._selection_model.selectedIndexes()
        if len(selected_indices) != 1:
            QtWidgets.QMessageBox.warning(self, 'Error', 'Please select a single track to split.')
            return
        
        selected_id = selected_indices[0].data(TrackListModel.IDRole)
        selected_start_frame = selected_indices[0].data(TrackListModel.StartFrameRole)
        
        split_idx = self._video_controller.get_frame_number() - selected_start_frame
        
        if not 1 <= split_idx <= len(selected_indices[0].data(TrackListModel.EventsRole)):
            QtWidgets.QMessageBox.warning(self, 'Error', 'Invalid split position for selected track.')
            return
        
        command = SplitTrackCommand(self._track_model, selected_id, split_idx)
        self._undo_stack.push(command)
        
        # Add the new track to the selection
        new_row_idx = self._track_model.index_by_id(command.new_track_id)
        self._selection_model.select(self._track_model.index(new_row_idx), QtCore.QItemSelectionModel.SelectionFlag.Select)
    
    def merge_selected_tracks(self):
        """Merge the selected tracks."""
        selected_indices = self._selection_model.selectedIndexes()
        if len(selected_indices) < 2:
            QtWidgets.QMessageBox.warning(self, 'Error', 'Please select at least two tracks to merge.')
            return
        
        selected_ids = [index.data(TrackListModel.IDRole) for index in selected_indices]
        
        command = MergeTracksCommand(self._track_model, selected_ids)
        self._undo_stack.push(command)
        
        # Add the new track to the selection
        new_row_idx = self._track_model.index_by_id(command.new_track_id)
        self._selection_model.select(self._track_model.index(new_row_idx), QtCore.QItemSelectionModel.SelectionFlag.Select)
    
    def seek_beginning_selected_tracks(self):
        """Seek to the first frame of the selected tracks."""
        selected_indices = self._selection_model.selectedIndexes()
        if len(selected_indices) < 1:
            QtWidgets.QMessageBox.warning(self, 'Error', 'Please select at least one track to seek to.')
            return
        
        selected_start_frames = [index.data(TrackListModel.StartFrameRole) for index in selected_indices]
        
        self._video_controller.set_frame(min(selected_start_frames))
    
    def seek_end_selected_tracks(self):
        """Seek to the last frame of the selected tracks."""
        selected_indices = self._selection_model.selectedIndexes()
        if len(selected_indices) < 1:
            QtWidgets.QMessageBox.warning(self, 'Error', 'Please select at least one track to seek to.')
            return
        
        selected_end_frames = [index.data(TrackListModel.EndFrameRole) for index in selected_indices]
        
        self._video_controller.set_frame(max(selected_end_frames))
    
    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self._video_controller.try_close()
        self._worker_thread.quit()
        return super().closeEvent(a0)



from PyQt6 import QtWidgets

from boxjelly.ui.settings.tabs.AbstractSettingsTab import AbstractSettingsTab


class VideoSettingsTab(AbstractSettingsTab):
    """
    Video settings tab.
    """
    
    def __init__(self, parent=None):
        super().__init__('Video', parent=parent)
        
        self._frame_rate_edit = QtWidgets.QDoubleSpinBox()
        self._frame_rate_edit.setRange(0.01, 1000.0)
        self._frame_rate_edit.setDecimals(2)
        self._frame_rate_edit.setSuffix(' fps')
        self._frame_rate_edit.setSingleStep(0.01)
        self._frame_rate_edit.setStepType(QtWidgets.QAbstractSpinBox.StepType.DefaultStepType)
        self._frame_rate_edit.setValue(self._settings.frame_rate.value)
        
        self._frame_rate_edit.valueChanged.connect(self._settings_changed)
        
        self._arrange()
    
    def _arrange(self):
        layout = QtWidgets.QFormLayout()
        
        layout.addRow('Frame Rate', self._frame_rate_edit)
        
        self.setLayout(layout)
    
    def apply_settings(self):
        self._settings.frame_rate.value = self._frame_rate_edit.value()
    


from PyQt6 import QtWidgets, QtCore, QtGui

from boxjelly.delegates.TrackHeadersDelegate import TrackHeadersDelegate


class TrackHeadersView(QtWidgets.QListView):
    def __init__(self, parent):
        super().__init__(parent)
        
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setItemDelegate(TrackHeadersDelegate(self))
        self.viewport().setMouseTracking(True)
        
        self.setFrameStyle(QtWidgets.QFrame.Shape.NoFrame)
        
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    
    def set_vertical_scroll(self, value: int):
        self.verticalScrollBar().setValue(value)
    
    def wheelEvent(self, e: QtGui.QWheelEvent):
        pass


"""
Application constants.
"""

# Application name
APP_NAME = 'BoxJelly'

# Application organization
APP_ORGANIZATION = 'MBARI'

# Network interface defaults
CONTROL_HOST = 'localhost'
CONTROL_PORT_DEFAULT = 5005
INCOMING_PORT_DEFAULT = 5562
INCOMING_TOPIC_DEFAULT = 'localization'
OUTGOING_PORT_DEFAULT = 5561
OUTGOING_TOPIC_DEFAULT = 'localization'

# Video frame rate
FRAME_RATE_DEFAULT = 29.97
